from ngSkinTools2.python_compatibility import Object
from ngSkinTools2.api import plugin, influenceMapping
from ngSkinTools2.api.layers import Layers
from ngSkinTools2.log import getLogger

log = getLogger("mirror")


class Mirror(Object):
    """
    query and configure mirror options for provided target
    """

    def __init__(self, target):
        """
        :type target: skin target (skinCluster or mesh)
        """
        self.target = target

    def seam_width(self):
        return plugin.ngst2Layers(self.target, q=True, mirrorWidth=True)

    def set_seam_width(self, seam_width):
        plugin.ngst2Layers(self.target, configureMirrorMapping=True, mirrorWidth=seam_width)

    def vertex_transfer_mode(self):
        return plugin.ngst2Layers(self.target, q=True, vertexTransferMode=True)

    def set_vertex_transfer_mode(self, mode):
        plugin.ngst2Layers(self.target, configureMirrorMapping=True, vertexTransferMode=mode)

    def axis(self):
        # type: () -> str
        """

        :return: 'x', 'y', 'z' if layers available, otherwise None
        """
        return plugin.ngst2Layers(self.target, q=True, mirrorAxis=True)

    def set_axis(self, axis):
        plugin.ngst2Layers(self.target, configureMirrorMapping=True, mirrorAxis=axis)
        self.recalculate_influences_mapping()

    def __mapper_config_attr(self):
        return Layers(self.target).data_node + ".influenceMappingOptions"

    def build_influences_mapper(self):
        from maya import cmds

        mapper = influenceMapping.InfluenceMapping()
        mapper.config.mirror_axis = self.axis()
        layers = Layers(self.target)
        mapper.influences = layers.list_influences()
        mapper.load_config_from_json(cmds.getAttr(self.__mapper_config_attr()))
        return mapper

    def save_influences_mapper(self, mapper):
        """
        :type mapper: influenceMapping.InfluenceMapping
        """
        from maya import cmds

        values = mapper.config_as_json()
        cmds.setAttr(self.__mapper_config_attr(), values, type='string')

    def set_influences_mapping(self, mapping):
        """
        :type mapping: map[int] -> int
        """
        log.info("mapping updated: %r", mapping)

        mapping_as_string = ','.join(str(k) + "," + str(v) for (k, v) in list(mapping.items()))
        plugin.ngst2Layers(self.target, configureMirrorMapping=True, influencesMapping=mapping_as_string)

    def recalculate_influences_mapping(self):
        """
        loads current influence mapping settings, and update influences mapping with these values
        """
        m = self.build_influences_mapper().calculate()
        self.set_influences_mapping(influenceMapping.InfluenceMapping.asIntIntMapping(m))

    def mirror(self, options):
        """
        :type options: MirrorOptions
        """
        plugin.ngst2Layers(
            self.target,
            mirrorLayerWeights=options.mirrorWeights,
            mirrorLayerMask=options.mirrorMask,
            mirrorLayerDq=options.mirrorDq,
            mirrorDirection=options.direction,
        )


class MirrorOptions(Object):
    directionNegativeToPositive = 0
    directionPositiveToNegative = 1
    directionGuess = 2
    directionFlip = 3

    def __init__(self):
        self.mirrorWeights = True
        self.mirrorMask = True
        self.mirrorDq = True
        self.direction = MirrorOptions.directionPositiveToNegative
